""" Installation of multi-file applications, as stored under src/apps """
import os
import shutil

from pathlib import Path

from .lib import insert
from . import package, disk

# Module methods
# ====================================================================


def get_rc_config(path):
    """provide config to be put in the rc files. Called by rc.compile_config_block()"""
    text = "# -- Apps\n"
    text += insert(get_fzf_rc_config(path))

    return text


# App: FZF
# ====================================================================


def get_fzf_rc_config(path):
    text = ""
    if "bash" in path.name:
        text = "# fzf\nsource ~/.rolv/fzf/.fzf.bash"
    if "zsh" in path.name:
        text = "# fzf\nsource ~/.rolv/fzf/.fzf.zsh"
    return text


def get_fzf_dst_path():
    return Path(os.getenv("HOME") + "/.rolv/fzf").resolve()


def install_fzf():
    disk.ensure_rolv_folder_exists()

    src_path = Path(package.get_src_path()).joinpath("apps/fzf")
    dst_path = get_fzf_dst_path()

    print(f"writing: {dst_path.as_posix()} (src: {src_path.as_posix()})")
    shutil.copytree(src_path, dst_path)


def uninstall_fzf():
    disk.rm_r(get_fzf_dst_path())


def sync_apps():
    uninstall_fzf()
    install_fzf()



"""
    Handles getting files that are packaged into the rolv page (rolv/src/*)
"""

import os
from pathlib import Path
from functools import cache
import importlib.util


def get_version():
    with open(get_path("version"), "r") as f:
        return f.read()


def get_src_path():
    path = importlib.util.find_spec("rolv.src").submodule_search_locations[0]
    return Path(path).resolve().as_posix()


@cache
def get_path(resource):
    path = importlib.util.find_spec("rolv.src").submodule_search_locations[0]
    return Path(os.path.join(path, resource))


@cache
def get_content(resource, encoding="utf-8"):
    path = GetIncludedResourcePath(resource)
    if encoding is False:
        with open(path, "rb") as f:
            return f.read()
    else:
        with open(path, "r", encoding=encoding) as f:
            return f.read()



"""
    Handles copying the files under src/executables to $HOME/.local/bin.
    Also removes old files from there, and makes sure the files are available via
    an alias.
"""

import os
import stat
from pathlib import Path

from . import package


# Module methods
# ====================================================================


def get_rc_config(path):
    """provide config to be put in the rc files. Called by rc.compile_config_block()"""
    lines = ["# -- Executables"]
    for executable in get_executable_paths():
        name = executable.stem
        lines.append(f'alias {name}="__rolv_{name}"')

    return "\n".join(lines)


# ====================================================================


def get_executable_paths():
    src_dir = Path(package.get_src_path())
    executables_dir = src_dir.joinpath("executables")
    executables = [f for f in executables_dir.iterdir() if f.is_file()]
    return executables


def sync_executables():
    bin_path = Path(os.getenv("HOME") + "/.local/bin").resolve()

    # remove existing execs starting with __rolv_
    for path in bin_path.iterdir():
        if path.name.startswith("__rolv_"):
            print(f"removing: {path.as_posix()}")
            os.remove(path)

    # place / overwrite executables
    for executable in get_executable_paths():
        if executable.name == "readme.md":
            continue

        with open(executable, "r") as f:
            content = f.read()

        dst = bin_path.joinpath(f"__rolv_{executable.name}")
        print(f"writing: {dst.as_posix()} (src: {executable.as_posix()})")

        with open(dst, "w") as f:
            f.write(content)

        make_executable(dst)


def make_executable(path):
    """Basically 'chmod +x path', but only for current user"""
    os.chmod(path, os.stat(path).st_mode | stat.S_IEXEC)



from . import rc, executables, package


def help_text():
    print(
        f"""
                  dP          
                  88          
88d888b. .d8888b. 88 dP   .dP 
88'  `88 88'  `88 88 88   d8' 
88       88.  .88 88 88 .88'  
dP       `88888P' dP 8888P'   v{package.get_version()}            
"""
    )
    print('"""A simple tool to synchronize all my \nquality of life scripts between servers."""\n')
    print("rolv.install    - install/update executables, aliases, etc")
    print("rolv.version    - get version of package")
    print("rolv.disclaimer - get a short list of things you need to know before using this package")
    print("\n")


def disclaimer():
    print(
        """
8888888b.  d8b                   888          d8b                                         
888  "Y88b Y8P                   888          Y8P                                         
888    888                       888                                                      
888    888 888 .d8888b   .d8888b 888  8888b.  888 88888b.d88b.   .d88b.  888d888 .d8888b  
888    888 888 88K      d88P"    888     "88b 888 888 "888 "88b d8P  Y8b 888P"   88K      
888    888 888 "Y8888b. 888      888 .d888888 888 888  888  888 88888888 888     "Y8888b. 
888  .d88P 888      X88 Y88b.    888 888  888 888 888  888  888 Y8b.     888          X88 
8888888P"  888  88888P'  "Y8888P 888 "Y888888 888 888  888  888  "Y8888  888      88888P' 

This package is written by me, for me. Best not to use this package at all if you are not me.

Will add/update the  ` # <rolv config> ... # </rolv config> `  block in any rc it can find.
  * Unless the ROLV_RC_FILE_PATH environment variable is set.
  
Will delete any file under  ` $HOME/.local/bin `  that starts with  ` __rolv_ `  .   
    """
    )


def version():
    print(package.get_version())


def install():
    print("\n")
    rc.set_rc_files()
    executables.sync_executables()
    apps.sync_apps()

    print("\nTo make sure the ENV is up to date, be sure to open a new terminal window, or run (one of) the following command(s):")
    for path in rc.get_rc_file_paths():
        print(f"  source {path.as_posix()}")
    print("\nDone.")
    print("\n")



"""
    This file handles all the interactions with the rc files (bashrc, zshrc, etc)
    It does not determine the config to be added itself, for that it calls get_rc_config() methods of other modules.
"""

import os
import re
from pathlib import Path

from . import executables, config, apps
from .lib import insert, trim_trailing_newlines


def get_rc_file_paths():
    """Finds rc file paths if they exist in the default location, unless ROLV_RC_FILE_PATH is set, then it will only return that value."""
    # use single, configured path
    explicit_path = os.getenv("ROLV_RC_FILE_PATH")
    if explicit_path is not None:
        explicit_path = Path(explicit_path).resolve()
        if explicit_path.exists() == False:
            print(f'Provided path ROLV_RC_FILE_PATH="{explicit_path.as_posix()}" does not exist. Get your shit together lmao. Aborting.')
            exit(1)
        return [Path(explicit_path).resolve()]

    # update all rc paths that are found
    res = []
    for path in [".zshrc", ".bashrc"]:
        abs_path = Path(os.getenv("HOME") + "/" + path).resolve()
        if abs_path.exists():
            res.append(abs_path)
    return res


def set_rc_files():
    """Adds/updates the config block to rc file paths."""
    for abs_path in get_rc_file_paths():
        set_rc_file(abs_path)


def compose_config(path):
    """Called to compose new config to be put in the bashrc. Lacks the fencing, just the config.
    If you want to read the current config block, see get_config_block()"""

    block = ""
    block += insert(config.get_rc_config(path))
    block += insert(executables.get_rc_config(path))
    block += insert(apps.get_rc_config(path))

    return block


def set_rc_file(path):
    """Updates/Places the rolv config block in the given rc file"""
    if path.exists() is False:
        print(f"File {path.as_posix()} not found. Aborting.")
        exit(0)

    rolv_block, other = get_config_block(path)
    if rolv_block:
        print(f"Rolv config block exists in {path.as_posix()} file - overwriting")
    else:
        print(f"Adding 'rolv config' block to {path.as_posix()}")

    contents = other

    # do nice spacing
    lines = [x.strip() for x in contents.split("\n")]
    if contents[-1] != "":
        contents += "\n"

    contents += "\n# <rolv config>\n# =======================================================\n"
    contents += insert(compose_config(path), space=0)
    contents += "# =======================================================\n# </rolv config>\n"

    with open(path, "w") as f:
        f.write(contents)


def remove_bashrc_block(path):
    block, other = get_config_block(path)
    with open(path, "w") as f:
        f.write(other)


def get_config_block(path, match_preceding_newlines=True):
    """Will return .bashrc content split into the rolv config block and all the other contents. To overwrite block, just use other content and add block to it"""

    with open(path, "r") as f:
        contents = f.read()

    in_block = False
    re_begin = re.compile(r"^ *# *<rolv config> *")
    re_end = re.compile(r"^ *# *</rolv config> *")

    other_lines = []
    block_lines = []
    for line in contents.split("\n"):
        # enter block
        if in_block == False:
            if re_begin.match(line):
                in_block = True

                # match single header empty newline with the matched block
                if match_preceding_newlines:
                    while len(other_lines) > 0 and other_lines[-1].strip() == "":
                        block_lines.append(other_lines.pop())

        # write all other lines to either block or "other" output
        if in_block:
            block_lines.append(line)
        else:
            other_lines.append(line)

        # exit block
        if in_block:
            if re_end.match(line):
                in_block = False

    # trim newlines at the end of the file
    other_lines = trim_trailing_newlines(other_lines)

    return "\n".join(block_lines), "\n".join(other_lines)



"""
    Provides content for rc.compile_config_block() based on content under
    src/config

    (Mostly hardcoded aliases/bash helper functions)
"""

import os
import stat
from pathlib import Path

from . import package
from .lib import insert, trim_newlines


# Module methods
# ====================================================================


def get_rc_config(path):
    """provide config to be put in the rc files. Called by rc.compile_config_block()"""
    src_dir = Path(package.get_src_path())

    block = ""

    for path in ["config/default_aliases_and_functions"]:
        abs_path = src_dir.joinpath(path)
        with open(abs_path, "r") as f:
            content = f.read()
        block += insert(trim_newlines(content))

    return block


# ====================================================================



""" Handles creating folders, removing folders, etc """

import os
import shutil
from pathlib import Path


def ensure_rolv_folder_exists():
    path = Path(os.getenv("HOME") + "/.rolv").resolve()
    path.mkdir(exist_ok=True)


def rm_r(path):
    if not os.path.exists(path):
        return
    if os.path.isfile(path) or os.path.islink(path):
        print(f"removing folder: {path.as_posix()}")
        os.unlink(path)
    else:
        print(f"removing folder: {path.as_posix()}")
        shutil.rmtree(path)



"""
    General helper functions
"""

import os


def insert(text, space=1):
    """Tests if last char is newline, and if not, will add the newline"""
    # make sure that the text at least ends on a new line
    if len(text) == 0 or text[-1] != "\n":
        text = text + "\n"

    # add extra spacing between blocks
    text += "\n" * space

    return text


def trim_trailing_newlines(text_or_lines):
    # make sure that we have lines as input
    if isinstance(text_or_lines, str):
        input_is_lines = False
        lines = text_or_lines.split("\n")
    else:
        input_is_lines = True
        lines = text_or_lines

    # remove trailing newlines
    while len(lines) > 0 and lines[-1].strip() == "":
        lines.pop()

    # return in some format as received
    if input_is_lines:
        return lines
    return "\n".join(lines)


def trim_preceeding_newlines(text_or_lines):
    # make sure that we have lines as input
    if isinstance(text_or_lines, str):
        input_is_lines = False
        lines = text_or_lines.split("\n")
    else:
        input_is_lines = True
        lines = text_or_lines

    # remove trailing newlines
    while len(lines) > 0 and lines[0].strip() == "":
        lines.pop(0)

    # return in some format as received
    if input_is_lines:
        return lines
    return "\n".join(lines)


def trim_newlines(text_or_lines):
    text_or_lines = trim_preceeding_newlines(text_or_lines)
    text_or_lines = trim_trailing_newlines(text_or_lines)
    return text_or_lines

